// Copyright 2011 Google Inc. All Rights Reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#include "msvc_helper.h"

#include "test.h"
#include "util.h"

TEST(EscapeForDepfileTest, SpacesInFilename) {
  ASSERT_EQ("sub\\some\\ sdk\\foo.h",
            EscapeForDepfile("sub\\some sdk\\foo.h"));
}

TEST(MSVCHelperTest, EnvBlock) {
  char env_block[] = "foo=bar\0";
  CLWrapper cl;
  cl.SetEnvBlock(env_block);
  string output;
  cl.Run("cmd /c \"echo foo is %foo%", &output);
  ASSERT_EQ("foo is bar\r\n", output);
}

TEST(MSVCHelperTest, NoReadOfStderr) {
  CLWrapper cl;
  string output;
  cl.Run("cmd /c \"echo to stdout&& echo to stderr 1>&2", &output);
  ASSERT_EQ("to stdout\r\n", output);
}
